# flake8: noqa

from __future__ import absolute_import, division, print_function, unicode_literals

from pkg_resources import get_distribution

from calico import __version__


def test_installed_version_should_match_package_version():
    assert get_distribution("calico").version == __version__



# Copyright (C) 2016-2019 H. Turgut Uyar <uyar@itu.edu.tr>
#
# Calico is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Calico is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calico.  If not, see <http://www.gnu.org/licenses/>.

"""Command line interface."""

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os
import sys
from argparse import ArgumentParser

from calico import __version__
from calico.parse import parse_spec


_logger = logging.getLogger("calico")

LOG_FILENAME = "calico.log"


def make_parser(prog):
    """Build a parser for command line arguments.

    :sig: (str) -> ArgumentParser
    :param prog: Name of program.
    :return: Created argument parser.
    """
    parser = ArgumentParser(prog=prog)
    parser.add_argument("--version", action="version", version="%(prog)s " + __version__)

    parser.add_argument("spec", help="test specifications file")
    parser.add_argument("-d", "--directory", help="change to directory before doing anything")
    parser.add_argument(
        "--validate", action="store_true", help="don't run tests, just validate spec"
    )
    parser.add_argument("-q", "--quiet", action="store_true", help="disable most messages")
    parser.add_argument("--log", action="store_true", help="log messages to file")
    parser.add_argument("--debug", action="store_true", help="enable debug messages")
    parser.add_argument("-t", "--tests", nargs="+", help="specify which tests cases will run")
    parser.add_argument(
        "--timeout", type=int, help="default timeout value for all test cases (seconds)"
    )
    return parser


def setup_logging(debug, log):
    """Set up logging levels and handlers.

    :sig: (bool, bool) -> None
    :param debug: Whether to activate debugging.
    :param log: Whether to log messages to a file.
    """
    _logger.setLevel(logging.DEBUG if debug else logging.INFO)

    # stream handler for console messages
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.DEBUG if debug else logging.INFO)
    _logger.addHandler(stream_handler)

    if log:
        # force debug mode
        _logger.setLevel(logging.DEBUG)

        # file handler for logging messages
        file_handler = logging.FileHandler(LOG_FILENAME)
        file_handler.setLevel(logging.DEBUG)
        _logger.addHandler(file_handler)


def main(argv=None):
    """Entry point of the utility.

    :sig: (Optional[List[str]]) -> None
    :param argv: Command line arguments.
    """
    argv = argv if argv is not None else sys.argv
    parser = make_parser(prog="calico")
    arguments = parser.parse_args(argv[1:])
    try:
        spec_filename = os.path.abspath(arguments.spec)
        with open(spec_filename) as f:
            content = f.read()

        if arguments.directory is not None:
            os.chdir(arguments.directory)

        setup_logging(debug=arguments.debug, log=arguments.log)

        runner = parse_spec(content)

        if not arguments.validate:
            report = runner.run(
                tests=arguments.tests, quiet=arguments.quiet, g_timeout=arguments.timeout
            )
            score = report["points"]
            print("Grade: %(s)s / %(p)s" % {"s": score, "p": runner.points})
    except Exception as e:
        print(e, file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright (C) 2019 H. Turgut Uyar <uyar@itu.edu.tr>
#
# Calico is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Calico is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calico.  If not, see <http://www.gnu.org/licenses/>.

"""General settings."""

from __future__ import absolute_import, division, print_function, unicode_literals

from distutils.spawn import find_executable


__version__ = "1.1.1"  # sig: str

GLOBAL_TIMEOUT = 2  # sig: int
"""Default timeout for tests, in seconds."""

SUPPORTS_JAIL = find_executable("fakechroot") is not None  # sig: bool
"""Whether this system supports changing root directory for a process."""



# THIS FILE IS AUTOMATICALLY GENERATED, DO NOT EDIT MANUALLY.

from typing import Any, Callable, Tuple

from ruamel.yaml import comments
from .base import Calico

SpecNode = comments.CommentedMap

def get_comment_value(node: SpecNode, name: str, field: str) -> str: ...
def get_attribute(
    node: SpecNode,
    test_name: str,
    names: Tuple[str],
    val_func: Callable[[Any, ...], bool],
    val_args: Any,
    err_message: str,
) -> Any: ...
def parse_spec(content: str) -> Calico: ...



from __future__ import absolute_import, division, print_function, unicode_literals

from pytest import raises

from calico.parse import parse_spec


def test_empty_spec_should_raise_error():
    with raises(AssertionError) as e:
        parse_spec("")
    assert "No test specification" in str(e)


def test_invalid_spec_format_should_raise_error():
    with raises(AssertionError) as e:
        parse_spec("!dummy")
    assert "Invalid test specification" in str(e)


def test_case_with_run_command_should_be_ok():
    source = """
      - c1:
          run: echo 1
    """
    runner = parse_spec(source)
    assert runner["c1"].command == "echo 1"


def test_case_with_run_shortcut_should_be_ok():
    source = """
      - c1:
          r: echo 1
    """
    runner = parse_spec(source)
    assert runner["c1"].command == "echo 1"


def test_case_order_should_be_preserved():
    source = """
      - c2:
          run: echo 2
      - c3:
          run: echo 3
      - c1:
          run: echo 1
    """
    runner = parse_spec(source)
    assert list(runner.keys()) == ["c2", "c3", "c1"]


def test_case_without_run_command_should_raise_error():
    source = """
      - c1:
          exit: 1
    """
    with raises(AssertionError) as e:
        parse_spec(source)
    assert "No run command" in str(e)


def test_case_with_multiple_run_commands_should_raise_error():
    source = """
      - c1:
          run:
            - echo 1a
            - echo 1b
    """
    with raises(AssertionError) as e:
        parse_spec(source)
    assert "Run command must be a string" in str(e)


def test_case_default_exit_status_should_be_zerd():
    source = """
      - c1:
          run: echo 1
    """
    runner = parse_spec(source)
    assert runner["c1"].exits == 0


def test_case_integer_exit_status_should_be_ok():
    source = """
      - c1:
          run: echo 1
          exit: 2
    """
    runner = parse_spec(source)
    assert runner["c1"].exits == 2


def test_case_fractional_exit_status_should_raise_error():
    source = """
      - c1:
          run: echo 1
          exit: 1.5
    """
    with raises(AssertionError) as e:
        parse_spec(source)
    assert "Exit status value must be an integer" in str(e)


def test_case_string_exit_status_should_raise_error():
    source = """
      - c1:
          run: echo 1
          exit: "0"
    """
    with raises(AssertionError) as e:
        parse_spec(source)
    assert "Exit status value must be an integer" in str(e)


def test_case_exit_shortcut_should_be_ok():
    source = """
      - c1:
          r: echo 1
          x: 2
    """
    runner = parse_spec(source)
    assert runner["c1"].exits == 2


def test_case_default_points_value_should_be_none():
    source = """
      - c1:
          run: echo 1
    """
    runner = parse_spec(source)
    assert runner["c1"].points is None


def test_case_integer_points_value_should_be_ok():
    source = """
      - c1:
          run: echo 1
          points: 10
    """
    runner = parse_spec(source)
    assert runner["c1"].points == 10


def test_case_fractional_points_value_should_be_ok():
    source = """
      - c1:
          run: echo 1
          points: 1.5
    """
    runner = parse_spec(source)
    assert runner["c1"].points == 1.5


def test_case_non_numeric_points_value_should_raise_error():
    source = """
      - c1:
          run: echo 1
          points: "10"
    """
    with raises(AssertionError) as e:
        parse_spec(source)
    assert "Points value must be numeric" in str(e)


def test_case_points_shortcut_should_be_ok():
    source = """
      - c1:
          run: echo 1
          p: 10
    """
    runner = parse_spec(source)
    assert runner["c1"].points == 10


def test_case_default_blocker_value_should_be_false():
    source = """
      - c1:
          run: echo 1
    """
    runner = parse_spec(source)
    assert not runner["c1"].blocker


def test_case_blocker_set_to_true_should_be_ok():
    source = """
      - c1:
          run: echo 1
          blocker: true
    """
    runner = parse_spec(source)
    assert runner["c1"].blocker


def test_case_blocker_set_to_false_should_be_ok():
    source = """
      - c1:
          run: echo 1
          blocker: false
    """
    runner = parse_spec(source)
    assert not runner["c1"].blocker


def test_case_non_boolean_blocker_value_should_raise_error():
    source = """
      - c1:
          run: echo 1
          blocker: maybe
    """
    with raises(AssertionError) as e:
        parse_spec(source)
    assert "Blocker value must be true or false" in str(e)


def test_case_blocker_shortcut_should_be_ok():
    source = """
      - c1:
          run: echo 1
          b: true
    """
    runner = parse_spec(source)
    assert runner["c1"].blocker


def test_case_default_visibility_value_should_be_true():
    source = """
      - c1:
          run: echo 1
    """
    runner = parse_spec(source)
    assert runner["c1"].visible


def test_case_visible_set_to_true_should_be_ok():
    source = """
      - c1:
          run: echo 1
          visible: true
    """
    runner = parse_spec(source)
    assert runner["c1"].visible


def test_case_visible_set_to_false_should_be_ok():
    source = """
      - c1:
          run: echo 1
          visible: false
    """
    runner = parse_spec(source)
    assert not runner["c1"].visible


def test_case_non_boolean_visibility_value_should_raise_error():
    source = """
      - c1:
          run: echo 1
          visible: maybe
    """
    with raises(AssertionError) as e:
        parse_spec(source)
    assert "Visibility value must be true or false" in str(e)


def test_case_visibility_shortcut_should_be_ok():
    source = """
      - c1:
          run: echo 1
          v: true
    """
    runner = parse_spec(source)
    assert runner["c1"].visible


def test_case_with_no_script_should_expect_eof():
    source = """
      - c1:
          run: echo 1
    """
    runner = parse_spec(source)
    assert [tuple(s) for s in runner["c1"].script] == [("e", "_EOF_", -1)]


def test_case_run_with_timeout_should_generate_expect_eof_with_timeout():
    source = """
      - c1:
          run: echo 1 # timeout: 5
    """
    runner = parse_spec(source)
    assert [tuple(s) for s in runner["c1"].script] == [("e", "_EOF_", 5)]


def test_case_run_with_non_numeric_timeout_value_should_raise_error():
    source = """
      - c1:
          run: echo 1 # timeout: "5"
    """
    with raises(AssertionError) as e:
        parse_spec(source)
    assert "Timeout value must be an integer" in str(e)


def test_case_script_with_invalid_action_should_raise_error():
    source = """
      - c1:
          run: echo 1
          script:
            - wait: 1
    """
    with raises(AssertionError) as e:
        parse_spec(source)
    assert "Unknown action type" in str(e)


def test_case_script_with_string_action_data_should_be_ok():
    source = """
      - c1:
          run: echo 1
          script:
            - expect: "1"
    """
    runner = parse_spec(source)
    assert [tuple(s) for s in runner["c1"].script] == [("e", "1", -1)]


def test_case_script_with_numeric_action_data_should_raise_error():
    source = """
      - c1:
          run: echo 1
          script:
            - expect: 1
    """
    with raises(AssertionError) as e:
        parse_spec(source)
    assert "Action data must be a string" in str(e)


def test_case_script_with_action_data_eof_should_be_ok():
    source = """
      - c1:
          run: echo 1
          script:
            - expect: _EOF_
    """
    runner = parse_spec(source)
    assert [tuple(s) for s in runner["c1"].script] == [("e", "_EOF_", -1)]


def test_case_script_with_multiple_action_data_should_raise_error():
    source = """
      - c1:
          run: echo 1
          script:
            - send:
                - "1a"
                - "1b"
    """
    with raises(AssertionError) as e:
        parse_spec(source)
    assert "Action data must be a string" in str(e)


def test_case_script_with_expect_shortcut_should_be_ok():
    source = """
      - c1:
          run: echo 1
          script:
            - e: "1"
    """
    runner = parse_spec(source)
    assert [tuple(s) for s in runner["c1"].script] == [("e", "1", -1)]


def test_case_script_with_send_shortcut_should_be_ok():
    source = """
      - c1:
          run: echo 1
          script:
            - s: "1"
    """
    runner = parse_spec(source)
    assert [tuple(s) for s in runner["c1"].script] == [("s", "1", -1)]


def test_case_script_order_should_be_preserved():
    source = """
      - c1:
          run: echo 1
          script:
            - expect: "foo"
            - send: "1"
            - expect: _EOF_
    """
    runner = parse_spec(source)
    assert [tuple(s) for s in runner["c1"].script] == [
        ("e", "foo", -1),
        ("s", "1", -1),
        ("e", "_EOF_", -1),
    ]


def test_case_script_action_with_integer_timeout_value_should_be_ok():
    source = """
      - c1:
          run: echo 1
          script:
            - expect: "foo" # timeout: 5
    """
    runner = parse_spec(source)
    assert [tuple(s) for s in runner["c1"].script] == [("e", "foo", 5)]


def test_case_script_action_with_fractional_timeout_value_should_raise_error():
    source = """
      - c1:
          run: echo 1
          script:
            - expect: "foo" # timeout: 4.5
    """
    with raises(AssertionError) as e:
        parse_spec(source)
    assert "Timeout value must be an integer" in str(e)


def test_case_script_action_with_string_timeout_value_should_raise_error():
    source = """
      - c1:
          run: echo 1
          script:
            - expect: "foo" # timeout: "5"
    """
    with raises(AssertionError) as e:
        parse_spec(source)
    assert "Timeout value must be an integer" in str(e)


def test_total_points_should_be_sum_of_points():
    source = """
      - c1:
          run: echo 1
          points: 15
      - c2:
          run: echo 2
      - c3:
          run: echo 3
          points: 25
    """
    runner = parse_spec(source)
    assert runner.points == 40


def test_total_fractional_points_should_be_sum_of_points():
    source = """
      - c1:
          run: echo 1
          points: 1.5
      - c2:
          run: echo 2
      - c3:
          run: echo 3
          points: 2.25
    """
    runner = parse_spec(source)
    assert runner.points == 3.75


def test_no_total_points_given_should_sum_zero():
    source = """
      - c1:
          run: echo 1
      - c2:
          run: echo 2
    """
    runner = parse_spec(source)
    assert runner.points == 0


def test_case_define_variable_should_be_ok():
    source = """
      - _define:
          vars:
            foo: bar
    """
    runner = parse_spec(source)
    assert runner["_define_vars"]["foo"] == "bar"



from __future__ import absolute_import, division, print_function, unicode_literals

from pytest import mark, raises

import os
import sys

PY2 = sys.version_info.major < 3

if not PY2:
    from unittest.mock import mock_open, patch

from pkg_resources import get_distribution

from calico import cli

base_dir = os.path.dirname(__file__)
circle_spec_file = os.path.join(base_dir, "circle.yaml")


def test_help_should_print_usage_and_exit(capsys):
    with raises(SystemExit):
        cli.main(argv=["calico", "--help"])
    out, err = capsys.readouterr()
    assert out.startswith("usage: ")


def test_cli_version_should_print_version_number_and_exit(capsys):
    version = get_distribution("calico").version
    with raises(SystemExit):
        cli.main(argv=["calico", "--version"])
    out, err = capsys.readouterr()
    assert out if not PY2 else err == "calico %(ver)s\n" % {"ver": version}


def test_no_spec_file_should_print_usage_and_exit(capsys):
    with raises(SystemExit):
        cli.main(argv=["calico"])
    out, err = capsys.readouterr()
    assert err.startswith("usage: ")
    assert ("required: spec" in err) or ("too few arguments" in err)


def test_existing_spec_file_should_be_ok(capsys):
    cli.main(argv=["calico", circle_spec_file])
    out, err = capsys.readouterr()
    assert err == ""


def test_non_existing_spec_file_should_exit_with_error(capsys):
    with raises(SystemExit):
        cli.main(argv=["calico", "dummy.spec"])
    out, err = capsys.readouterr()
    assert "No such file or directory:" in err


# TODO: add tests for -d option


def test_non_existing_base_directory_should_exit_with_error(capsys):
    with raises(SystemExit):
        cli.main(argv=["calico", "-d", "dummy", circle_spec_file])
    out, err = capsys.readouterr()
    assert "No such file or directory:" in err


def test_validate_valid_spec_file_should_not_print_output(capsys):
    cli.main(argv=["calico", "--validate", circle_spec_file])
    out, err = capsys.readouterr()
    assert out == ""


@mark.skipif(PY2, reason="has to mock builtins.open")
def test_validate_invalid_spec_file_should_exit_with_error(capsys):
    with patch("builtins.open", mock_open(read_data=""), create=True):
        with raises(SystemExit):
            cli.main(argv=["calico", "--validate", circle_spec_file])
        out, err = capsys.readouterr()
        assert "No test specification" in err


# TODO: add tests for --quiet option


# TODO: add tests for --log option


# TODO: add tests for --debug option


# TODO: add tests for summary output



from __future__ import absolute_import, division, print_function, unicode_literals

from calico.base import Action, ActionType, run_script


def test_script_expect_eof_should_be_ok():
    result = run_script("true", [Action(ActionType.EXPECT, "_EOF_")])
    assert result == (0, [])


def test_script_empty_should_be_ok():
    result = run_script("true", [])
    assert result == (0, [])


def test_script_expect_output_should_be_ok():
    result = run_script(
        "echo 1", [Action(ActionType.EXPECT, "1"), Action(ActionType.EXPECT, "_EOF_")]
    )
    assert result == (0, [])


def test_script_closing_expect_feof_should_be_optional():
    result = run_script("echo 1", [Action(ActionType.EXPECT, "1")])
    assert result == (0, [])


def test_script_expect_with_timeout_should_be_ok():
    result = run_script("sleep 1", [Action(ActionType.EXPECT, "_EOF_", timeout=2)])
    assert result == (0, [])


def test_script_expect_with_exceeded_timeout_should_report_error():
    result = run_script("sleep 2", [Action(ActionType.EXPECT, "_EOF_", timeout=1)])
    assert result == (None, ["Timeout exceeded."])


def test_script_expect_with_unmatched_output_should_report_error():
    result = run_script("true", [Action(ActionType.EXPECT, "1")])
    assert result == (0, ["Expected output not received."])


# def test_script_expect_with_extra_output_should_report_error():
#     result = run_script("echo 1", [Action(ActionType.EXPECT, "_EOF_")])
#     assert result == (0, ["Extra output received."])


def test_script_send_input_should_be_ok():
    result = run_script(
        "bash -c 'read x && echo $x'",
        [Action(ActionType.SEND, "1"), Action(ActionType.EXPECT, "1")],
    )
    assert result == (0, [])


def test_timeout_should_kill_infinite_program():
    result = run_script("yes", [Action(ActionType.EXPECT, "_EOF_", timeout=1)])
    assert result == (None, ["Timeout exceeded."])



# THIS FILE IS AUTOMATICALLY GENERATED, DO NOT EDIT MANUALLY.

from typing import Any, List, Mapping, Optional, Tuple, Union

from collections import OrderedDict
from enum import Enum

PY2 = ...  # type: str


class ActionType(Enum):
    EXPECT = ...  # type: Tuple[str, str]
    SEND = ...  # type: Tuple[str, str]

class Action:
    type_ = ...  # type: ActionType
    data = ...  # type: str
    timeout = ...  # type: Optional[int]
    def __init__(
        self, type_: ActionType, data: str, timeout: Optional[int] = ...
    ) -> None: ...

def run_script(
    command: str,
    script: List[Action],
    defs: Optional[Mapping] = ...,
    g_timeout: Optional[int] = ...,
) -> Tuple[int, List[str]]: ...

class TestCase:
    name = ...  # type: str
    command = ...  # type: str
    script = ...  # type: List[Action]
    timeout = ...  # type: Optional[int]
    exits = ...  # type: Optional[int]
    points = ...  # type: Optional[Union[int, float]]
    blocker = ...  # type: bool
    visible = ...  # type: bool
    def __init__(
        self,
        name: str,
        command: str,
        timeout: Optional[int] = ...,
        exits: Optional[int] = ...,
        points: Optional[Union[int, float]] = ...,
        blocker: Optional[bool] = ...,
        visible: Optional[bool] = ...,
    ) -> None: ...
    def add_action(self, action: Action) -> None: ...
    def run(
        self,
        defs: Optional[Mapping] = ...,
        jailed: Optional[bool] = ...,
        g_timeout: Optional[int] = ...,
    ) -> Mapping[str, Union[str, List[str]]]: ...

class Calico(OrderedDict):
    points = ...  # type: Union[int, float]
    def __init__(self) -> None: ...
    def add_case(self, case: TestCase) -> None: ...
    def run(
        self,
        tests: Optional[bool] = ...,
        quiet: Optional[List[str]] = ...,
        g_timeout: Optional[int] = ...,
    ) -> Mapping[str, Any]: ...



# THIS FILE IS AUTOMATICALLY GENERATED, DO NOT EDIT MANUALLY.

from typing import List, Optional

from argparse import ArgumentParser

def make_parser(prog: str) -> ArgumentParser: ...
def setup_logging(debug: bool, log: bool) -> None: ...
def main(argv: Optional[List[str]] = ...) -> None: ...



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['calico']

package_data = \
{'': ['*']}

install_requires = \
['enum34>=1.1,<2.0', 'pexpect>=4.6,<5.0', 'ruamel.yaml>=0.15.41,<0.16.0']

entry_points = \
{'console_scripts': ['calico = calico.cli:main']}

setup_kwargs = {
    'name': 'calico',
    'version': '1.1.1',
    'description': 'I/O checker for command line programs.',
    'long_description': 'Calico is a utility for checking command-line programs in terms of their\ninput and output. It checks whether a program generates the correct output\nwhen given some inputs. It was developed to evaluate simple programming\nassignments in an introductory programming course.\n\nGetting started\n---------------\n\nYou can install Calico using pip::\n\n   pip install calico\n\nCalico uses `pexpect`_ for interacting with the program it is checking.\nThe file that specifies the inputs and outputs for the checks\nis in `YAML`_ format.\n\n.. _pexpect: https://pexpect.readthedocs.io/\n.. _YAML: http://www.yaml.org/\n\nGetting help\n------------\n\nThe online documentation is available on: https://calico.readthedocs.io/\n\nThe source code can be obtained from: https://github.com/itublg/calico\n\nLicense\n-------\n\nCopyright (C) 2016-2019 H. Turgut Uyar <uyar@itu.edu.tr>\n\nSee ``AUTHORS.rst`` for a list of all contributors.\n\nCalico is released under the GPL license, version 3 or later. Read\nthe included ``LICENSE.txt`` for details.\n',
    'author': 'H. Turgut Uyar',
    'author_email': 'uyar@itu.edu.tr',
    'url': 'https://github.com/itublg/calico',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
}


setup(**setup_kwargs)



# Copyright (C) 2016-2019 H. Turgut Uyar <uyar@itu.edu.tr>
#
# Calico is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Calico is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calico.  If not, see <http://www.gnu.org/licenses/>.

"""Base classes for Calico."""

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os
import sys
from collections import OrderedDict
from enum import Enum

import pexpect

from . import GLOBAL_TIMEOUT, SUPPORTS_JAIL


PY2 = sys.version_info < (3,)  # sig: str

MAX_LEN = 40

_logger = logging.getLogger("calico")


class ActionType(Enum):
    """Type of an action."""

    EXPECT = ("e", "expect")  # sig: Tuple[str, str]
    SEND = ("s", "send")  # sig: Tuple[str, str]


class Action:
    """An action in a test script."""

    def __init__(self, type_, data, timeout=-1):
        """Initialize this action.

        :sig: (ActionType, str, Optional[int]) -> None
        :param type_: Expect or send.
        :param data: Data to expect or send.
        :param timeout: Timeout duration, in seconds.
        """
        self.type_ = type_  # sig: ActionType
        """Type of this action, expect or send."""

        self.data = data if data != "_EOF_" else pexpect.EOF  # sig: str
        """Data description of this action, what to expect or send."""

        self.timeout = timeout  # sig: Optional[int]
        """Timeout duration of this action."""

    def __iter__(self):
        """Get components of this action as a sequence."""
        yield self.type_.value[0]
        yield self.data if self.data != pexpect.EOF else "_EOF_"
        yield self.timeout


def run_script(command, script, defs=None, g_timeout=None):
    """Run a command and check whether it follows a script.

    :sig: (str, List[Action], Optional[Mapping], Optional[int]) -> Tuple[int, List[str]]
    :param command: Command to run.
    :param script: Script to check against.
    :param defs: Variable substitutions.
    :param g_timeout: Global timeout value for the spawn class
    :return: Exit status and errors.
    """
    defs = defs if defs is not None else {}
    g_timeout = g_timeout if g_timeout is not None else GLOBAL_TIMEOUT

    process = pexpect.spawn(command, timeout=g_timeout)
    process.setecho(False)
    errors = []

    last = script[-1] if len(script) > 0 else None
    if (last is None) or ((last.type_ != ActionType.EXPECT) and (last.data != "_EOF_")):
        script.append(Action(ActionType.EXPECT, "_EOF_"))

    for action in script:
        if action.data is not pexpect.EOF:
            action.data = action.data % defs
        if action.type_ == ActionType.EXPECT:
            try:
                expecting = (
                    "_EOF_" if action.data is pexpect.EOF else ('"%(a)s"' % {"a": action.data})
                )
                timeout = action.timeout if action.timeout != -1 else g_timeout
                _logger.debug("  expecting (%ds): %s", timeout, expecting)
                process.expect(action.data, timeout=action.timeout)
                output = process.after
                received = (
                    "_EOF_" if ".EOF" in repr(output) else ('"%(o)s"' % {"o": output.decode()})
                )
                _logger.debug("  received: %s", received)
            except pexpect.EOF:
                output = process.before
                received = (
                    "_EOF_" if ".EOF" in repr(output) else ('"%(o)s"' % {"o": output.decode()})
                )
                _logger.debug('  received: "%s"', received)
                process.close(force=True)
                _logger.debug("FAILED: Expected output not received.")
                errors.append("Expected output not received.")
                break
            except pexpect.TIMEOUT:
                output = process.before
                received = (
                    "_EOF_" if ".EOF" in repr(output) else ('"%(o)s"' % {"o": output.decode()})
                )
                _logger.debug('  received: "%s"', received)
                process.close(force=True)
                _logger.debug("FAILED: Timeout exceeded.")
                errors.append("Timeout exceeded.")
                break
        elif action.type_ == ActionType.SEND:
            _logger.debug('  sending: "%s"', action.data)
            process.sendline(action.data)
    else:
        process.close(force=True)
    return process.exitstatus, errors


class TestCase:
    """A case in a test suite."""

    def __init__(
        self, name, command, timeout=-1, exits=0, points=None, blocker=False, visible=True
    ):
        """Initialize this test case.

        :sig:
            (
                str,
                str,
                Optional[int],
                Optional[int],
                Optional[Union[int, float]],
                Optional[bool],
                Optional[bool]
            ) -> None
        :param name: Name of the case.
        :param command: Command to run.
        :param timeout: Timeout duration, in seconds.
        :param exits: Expected exit status.
        :param points: Contribution to overall points.
        :param blocker: Whether failure blocks subsequent cases.
        :param visible: Whether the test will be visible during the run.
        """
        self.name = name  # sig: str
        """Name of this test case."""

        self.command = command  # sig: str
        """Command to run in this test case."""

        self.script = []  # sig: List[Action]
        """Sequence of actions to run in this test case."""

        self.timeout = timeout  # sig: Optional[int]
        """Timeout duration of this test case, in seconds."""

        self.exits = exits  # sig: Optional[int]
        """Expected exit status of this test case."""

        self.points = points  # sig: Optional[Union[int, float]]
        """How much this test case contributes to the total points."""

        self.blocker = blocker  # sig: bool
        """Whether failure in this case will block subsequent cases or not."""

        self.visible = visible  # sig: bool
        """Whether this test will be visible during the run or not."""

    def add_action(self, action):
        """Append an action to the script of this test case.

        :sig: (Action) -> None
        :param action: Action to append to the script.
        """
        self.script.append(action)

    def run(self, defs=None, jailed=False, g_timeout=None):
        """Run this test and produce a report.

        :sig:
            (
                Optional[Mapping],
                Optional[bool],
                Optional[int]
            ) -> Mapping[str, Union[str, List[str]]]
        :param defs: Variable substitutions.
        :param jailed: Whether to jail the command to the current directory.
        :param g_timeout: Global timeout for all expects in the test
        :return: Result report of the test.
        """
        report = {"errors": []}

        jail_prefix = ("fakechroot chroot %(d)s " % {"d": os.getcwd()}) if jailed else ""
        command = "%(j)s%(c)s" % {"j": jail_prefix, "c": self.command}
        _logger.debug("running command: %s", command)

        exit_status, errors = run_script(
            self.command, self.script, defs=defs, g_timeout=g_timeout
        )
        report["errors"].extend(errors)

        _logger.debug("exit status: %d (expected %d)", exit_status, self.exits)
        if exit_status != self.exits:
            report["errors"].append("Incorrect exit status.")

        return report


class Calico(OrderedDict):
    """A suite containing a collection of ordered test cases."""

    def __init__(self):
        """Initialize this test suite from a given specification.

        :sig: () -> None
        """
        if PY2:
            OrderedDict.__init__(self)
        else:
            super().__init__()

        self.points = 0  # sig: Union[int, float]
        """Total points in this test suite."""

    def add_case(self, case):
        """Add a test case to this suite.

        :sig: (TestCase) -> None
        :param case: Test case to add.
        """
        if PY2:
            OrderedDict.__setitem__(self, case.name, case)
        else:
            super().__setitem__(case.name, case)
        self.points += case.points if case.points is not None else 0

    def run(self, tests=None, quiet=False, g_timeout=None):
        """Run this test suite.

        :sig: (Optional[bool], Optional[List[str]], Optional[int]) -> Mapping[str, Any]
        :param tests: Tests to include in the run.
        :param quiet: Whether to suppress progress messages.
        :param g_timeout: Global timeout value for the all tests
        :return: A report containing the results.
        """
        report = OrderedDict()
        earned_points = 0

        os.environ["TERM"] = "dumb"  # disable color output in terminal

        test_names = tests if tests is not None else [n for n in self.keys() if n[0] != "_"]
        for test_name in test_names:
            test = self.get(test_name)

            _logger.debug("starting test %s", test_name)
            if (not quiet) and test.visible:
                dots = "." * (MAX_LEN - len(test_name) + 1)
                print("%(t)s %(d)s" % {"t": test_name, "d": dots}, end=" ")

            jailed = SUPPORTS_JAIL and test_name.startswith("case_")
            report[test_name] = test.run(
                defs=self.get("_define_vars"), jailed=jailed, g_timeout=g_timeout
            )
            passed = len(report[test_name]["errors"]) == 0

            if test.points is None:
                if (not quiet) and test.visible:
                    print("PASSED" if passed else "FAILED")
            else:
                report[test_name]["points"] = test.points if passed else 0
                earned_points += report[test_name]["points"]
                if (not quiet) and test.visible:
                    scored = report[test_name]["points"]
                    print("%(s)s / %(p)s" % {"s": scored, "p": test.points})

            if test.blocker and (not passed):
                break

        report["points"] = earned_points
        return report



# Copyright (C) 2016-2019 H. Turgut Uyar <uyar@itu.edu.tr>
#
# Calico is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Calico is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calico.  If not, see <http://www.gnu.org/licenses/>.

"""Specification parsing."""

from __future__ import absolute_import, division, print_function, unicode_literals

from ruamel import yaml
from ruamel.yaml import comments

from .base import Action, ActionType, Calico, TestCase


# sigalias: SpecNode = comments.CommentedMap


def get_comment_value(node, name, field):
    """Get the value of a comment field.

    :sig: (SpecNode, str, str) -> str
    :param node: Node to get the comment from.
    :param name: Name of setting in the node.
    :param field: Name of comment field.
    :return: Value of comment field.
    """
    try:
        comment = node.ca.items[name][2].value[1:].strip()  # remove the leading hash
    except KeyError:
        comment = None
    if comment is not None:
        delim = field + ":"
        if comment.startswith(delim):
            return comment[len(delim) :].strip()
    return None


def get_attribute(node, test_name, names, val_func, val_args, err_message):
    """Get the value of an test attribute.

    :sig: (SpecNode, str, Tuple[str], Callable[[Any, ...], bool], Any, str) -> Any
    :param node: Node to get the attribute
    :param test_name: Name of the test
    :param names: Long and short names of the attribute
    :param val_func: A validator function for the attribute
    :param val_args: An argument to the validator function
    :param err_message: An error message to shown when validation fails
    """
    _short, _long = names
    attr = node.get(_long, node.get(_short))
    if attr is not None:
        if val_args is None:
            result = val_func(attr)
        else:
            result = val_func(attr, val_args)

        assert result, err_message % test_name
    return attr


def parse_spec(content):
    """Parse a test specification.

    :sig: (str) -> Calico
    :param content: Specification to parse.
    :return: Created Calico runner.
    :raise AssertionError: When given specification is invalid.
    """
    try:
        spec = yaml.round_trip_load(content)
    except yaml.YAMLError as e:
        raise AssertionError(str(e))

    if spec is None:
        raise AssertionError("No test specification")

    if not isinstance(spec, comments.CommentedSeq):
        raise AssertionError("Invalid test specification")

    action_types = {i: m for m in ActionType for i in m.value}

    runner = Calico()

    tests = [(n, t) for c in spec for n, t in c.items()]
    attributes = [
        (
            "command",
            {
                "names": ("r", "run"),
                "val_func": isinstance,
                "val_args": str,
                "err_message": "%s: Run command must be a string",
            },
        ),
        (
            "points",
            {
                "names": ("p", "points"),
                "val_func": isinstance,
                "val_args": (int, float),
                "err_message": "%s: Points value must be numeric",
            },
        ),
        (
            "blocker",
            {
                "names": ("b", "blocker"),
                "val_func": isinstance,
                "val_args": bool,
                "err_message": "%s: Blocker value must be true or false",
            },
        ),
        (
            "exits",
            {
                "names": ("x", "exit"),
                "val_func": isinstance,
                "val_args": int,
                "err_message": "%s: Exit status value must be an integer",
            },
        ),
        (
            "visible",
            {
                "names": ("v", "visible"),
                "val_func": isinstance,
                "val_args": bool,
                "err_message": "%s: Visibility value must be true or false",
            },
        ),
    ]

    for test_name, test in tests:
        if test_name[0] == "_":
            for section, section_value in test.items():
                runner[test_name + "_" + section] = section_value
            continue

        kwargs = {}
        for kwarg, attr in attributes:
            attr_ = get_attribute(test, test_name, **attr)
            if attr_ is not None:
                kwargs[kwarg] = attr_

        assert "command" in kwargs, "%(t)s: No run command" % {"t": test_name}

        timeout = get_comment_value(test, name="run", field="timeout")
        if timeout is not None:
            assert timeout.isdigit(), "%(t)s: Timeout value must be an integer" % {
                "t": test_name
            }
            kwargs["timeout"] = int(timeout)

        case = TestCase(test_name, **kwargs)

        script = test.get("script")
        if script is None:
            # If there's no script, just expect EOF.
            action = Action(ActionType.EXPECT, "_EOF_", timeout=case.timeout)
            case.add_action(action)
        else:
            for step in script:
                action_type, data = [(k, v) for k, v in step.items()][0]
                assert action_type in action_types, "%(t)s: Unknown action type" % {
                    "t": test_name
                }
                assert isinstance(data, str), "%(t)s: Action data must be a string" % {
                    "t": test_name
                }

                kwargs = {}

                timeout = get_comment_value(step, name=action_type, field="timeout")
                if timeout is not None:
                    assert timeout.isdigit(), "%(t)s: Timeout value must be an integer" % {
                        "t": test_name
                    }
                    kwargs["timeout"] = int(timeout)

                action = Action(action_types[action_type], data, **kwargs)
                case.add_action(action)

        runner.add_case(case)

    return runner
